/* 
 * John David Ratliff
 * jdratlif@cs.indiana.edu
 * Monday 2:30 - 3:20 Lab
 */

package scheme.interp;

import scheme.util.*;
import scheme.absyn.*;

public class Eval_Exp implements ExpVisitor {
    private SchemeValue v;
    private Eval_ExpList eval_explist;
    private Eval_Constant eval_const;
    private Eval_Datum eval_datum;
    private Env env;
    
    // added a Store to the interpreter
    private Store s;

    public SchemeValue getValue() { return v; }

    public Env getEnv() { return env; }
    
    // added method to get current Store
    public Store getStore() { return this.s; }

    public void setEnv (Env newenv) { env = newenv; }

    public Eval_Exp () {
        this.eval_const = new Eval_Constant();
        this.eval_datum = new Eval_Datum(eval_const);
        this.eval_explist = new Eval_ExpList(this);
        
        // changed to use new version of Environment
        this.env = new Env();
        
        // added Store initialization
        this.s = new Store();
    }

    public void visitConstantE (ConstantE e) throws VisitE {
        e.get().accept(eval_const);
        v = eval_const.getValue();
    }

    public void visitQuoteE (QuoteE e) throws VisitE {
        e.get().accept(eval_datum);
        v = eval_datum.getValue();
    }

    public void visitVarE (VarE e) throws VisitE {
        // lookup the value of the variable in the environment
        int location = ((Integer)this.env.find(e.getName())).intValue();        
        this.v = this.s.find(location);
    }

    public void visitIfE (IfE e) throws VisitE {
        // evaluate the test
        e.getTest().accept(this);
        
        if (v.isTrue()) {
            // if true, evaluate the true-expression
            e.getThen().accept(this);
        } else {
            // otherwise, evaluate the false-expression
            e.getElse().accept(this);
        }
    }

    public void visitLambdaE (LambdaE e) throws VisitE {
       this.v = new ClosureV(e.getIds(),e.getBody(),getEnv());
    }

    public void visitAppE (AppE e) throws VisitE {
        ClosureV c;
        
        // evalute the function name
        e.getFun().accept(this);
        
        try {
            // get the closure
            c = (ClosureV)v;
            
            // evaluate the arguments
            e.getArgs().accept(this.eval_explist);
            
            // apply the function with the arguments
            this.v = c.apply(eval_explist.getValues(),this);
        } catch (ClassCastException ex) {
            throw new VisitE("trying to apply a non-function");
        }
    } 

	// changed to use Store/Environment combo
    public void visitSetE (SetE e) throws VisitE {
        // evalute the expression
        e.getExp().accept(this);

		// Store location in environment. Value is in Store.
        int location = ((Integer)this.env.find(e.getId())).intValue();
        this.s.update(location,v);

        // 'return' the set! expression value
        v = new VoidV();
    }

    public void visitBeginE (BeginE e) throws VisitE {
        e.getExps().accept(eval_explist);
        v = (SchemeValue)eval_explist.getValues().getLast();
    }

    public void visitPrimappE (PrimappE e) throws VisitE {
        e.getArgs().accept(eval_explist);
        this.v = applyPrim(e.getOp(),eval_explist.getValues());
    }

	// modified to use Store/Environment combo
    public void visitLetE (LetE e) throws VisitE {
        List xs = e.getDecls();
        Decl d;
        Env newenv = getEnv(), safe = getEnv();
        int location;

        while (!(xs instanceof EmptyL)) {
            // get the next Declaration
            d = (Decl)xs.hd();
            
            // evalute the expression
            d.getExp().accept(this);
            
            // add item to the store
            location = this.s.alloc(v);
            
            // add location in store to environment
            newenv = newenv.add(d.getVar(), new Integer(location));
            
            xs = xs.tl();
        }

        // setup the new environment
        setEnv(newenv);
        
        // now evalutate the body
        e.getBody().accept(this);
        
        // environment is only valid inside body of let
        setEnv(safe);
    }

	// modified to use Store/Environment combo
    public void visitLetrecE (LetrecE e) throws VisitE {
        List ds = e.getDecls();
        Decl d;
        Env before = getEnv();
        int location;
        
        while (!(ds instanceof EmptyL)) {
            // get the next Declaration
            d = (Decl)ds.hd();
            
            // add item to the store
            location = this.s.alloc(new VoidV());
            
            // add location in store to environment
            this.env = this.env.add(d.getVar(), new Integer(location));
            
            ds = ds.tl();
        }
        
        ds = e.getDecls();
        
        while (!(ds instanceof EmptyL)) {
            ((Decl)ds.hd()).getExp().accept(this);
            
            location = ((Integer)this.env.find(((Decl)ds.hd()).getVar())).intValue();
            this.s.update(location,this.v);
            
            ds = ds.tl();
        }
        
        // now evaluate the body of the letrec in the recursive environment
        e.getBody().accept(this);
        
        // restore the previous environment
        setEnv(before);
    }

    private SchemeValue applyPrim (byte op, List vs) throws VisitE {
        SchemeValue v;
        
        try {
            // match against the primitives
            switch (op) {
                // (add1 value)
                case Prim.ADD1:
                    v = new IntV(((IntV)vs.hd()).getInt() + 1);
                    break;
                case Prim.BOOLEANP:
                    v = new BoolV(vs.hd() instanceof BoolV);
                    break;
                case Prim.CAR:
                    v = ((PairV)vs.hd()).getCar();
                    break;                    
                case Prim.CDR:
                    v = ((PairV)vs.hd()).getCdr();
                    break;
                case Prim.CONS:
                    v = new PairV((SchemeValue)vs.hd(),(SchemeValue)vs.tl().hd());
                    break;
                case Prim.EQ:
                    v = new BoolV(((IntV)vs.hd()).getInt() == ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.EQP:
                    v = new BoolV(vs.hd() == vs.tl().hd());
                    break;
                case Prim.INTEGERP:
                    v = new BoolV(vs.hd() instanceof IntV);
                    break;
                case Prim.LEQ:
                    v = new BoolV(((IntV)vs.hd()).getInt() <= ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.LESS:
                    v = new BoolV(((IntV)vs.hd()).getInt() < ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.MINUS:
                    v = new IntV(((IntV)vs.hd()).getInt() - ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.NOT:
                    v = new BoolV(!((SchemeValue)vs.hd()).isTrue());
                    break;
                case Prim.NULLP:
                    v = new BoolV(vs.hd() instanceof EmptyV);
                    break;
                case Prim.PAIRP:
                    v = new BoolV(vs.hd() instanceof PairV);
                    break;
                case Prim.PLUS:
                    v = new IntV(((IntV)vs.hd()).getInt() + ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.PROCEDUREP:
                    v = new BoolV(vs.hd() instanceof ClosureV);
                    break;
                case Prim.SETCAR:
                    ((PairV)vs.hd()).setCar((SchemeValue)vs.tl().hd());
                    v = new VoidV();
                    break;
                case Prim.SETCDR:
                    ((PairV)vs.hd()).setCdr((SchemeValue)vs.tl().hd());
                    v = new VoidV();
                    break;
                case Prim.SUB1:
                    v = new IntV(((IntV)vs.hd()).getInt() - 1);
                    break;
                case Prim.TIMES:
                    v = new IntV(((IntV)vs.hd()).getInt() * ((IntV)vs.tl().hd()).getInt());
                    break;
                case Prim.VOID:
                    v = new VoidV();
                    break;
                case Prim.ZEROP:
                    v = new BoolV(((IntV)vs.hd()).getInt() == 0);
                    break;
                    
                // anything else is invalid
                default: throw new VisitE("invalid primitive operation");
            }
        } catch (ClassCastException ex) {
            throw new VisitE("invalid argument to primitive");
        } catch (EmptyListE ex) {
            throw new VisitE("invalid number of arguments to primitive");
        }
        
        return v;
    }
}
